#include "biogeo.h"
#include "voronoi.h"

static SFLAG CompareSets (BITFIELD* nodeAsetA, BITFIELD* nodeAsetB, BITFIELD* nodeBsetA, BITFIELD* nodeBsetB);
static SFLAG IsNodeComp (PHYLONODE* nodeA, PHYLONODE* nodeB);

void DestroyAlloSet (ALLOPATRY* alloSet) {
 if (alloSet -> triangles != NULL)
   DestroyTriangleList (alloSet -> triangles);
 g_list_free (alloSet -> nodeList);
 g_slice_free1 (GETBITFIELDSIZE, alloSet -> setA);
 g_slice_free1 (GETBITFIELDSIZE, alloSet -> setB);
 g_slice_free1 (GETBITFIELDSIZE, alloSet -> baseA);
 g_slice_free1 (GETBITFIELDSIZE, alloSet -> baseB);
}

SFLAG IsNodeComp (PHYLONODE* nodeA, PHYLONODE* nodeB) {
 SFLAG toRet;

 toRet = TRUE;
 if (nodeA -> tree == nodeB -> tree) {
   if (IsDesc (nodeA -> node, nodeB -> node))
     toRet = FALSE;
   else if (IsDesc (nodeB -> node, nodeA -> node))
     toRet = FALSE;
 }
 return toRet;
}

SFLAG CompareSets (BITFIELD* nodeAsetA, BITFIELD* nodeAsetB, BITFIELD* nodeBsetA, BITFIELD* nodeBsetB) {
 SFLAG toRet;

 toRet = FALSE;
 if (BitFieldAnd (nodeAsetA, nodeBsetA)) {
   if (BitFieldAnd (nodeAsetB, nodeBsetB) && (!(BitFieldAnd (nodeAsetA, nodeBsetB))) && (!(BitFieldAnd (nodeAsetB, nodeBsetA))))
     toRet = TRUE;
 }
 else if (BitFieldAnd (nodeAsetA, nodeBsetB)) {
   if (BitFieldAnd (nodeAsetB, nodeBsetA) && (!(BitFieldAnd (nodeAsetA, nodeBsetA))) && (!(BitFieldAnd (nodeAsetB, nodeBsetB))))
     toRet = TRUE;
 }
 return toRet;
}

void MakeCommonBarriers () {
 GList* navTree;
 GList* nuList;
 GList* navComp;
 GNode* navNode;
 PHYLOTREE* tree;
 PHYLONODE* node;
 PHYLONODE* compNode;
 SCOUNTER numNodes, j, k;
 GPtrArray* nodeArray;
 BITFIELD* tmpField;

 /* Checks */
 tree = GETTREES -> data;
 if (tree -> recCons != NULL) {
   MakeConsensus ();
   return;
 }
 RemoveAlloSets ();
 MakeConsensus ();

 /* Initialization */
 numNodes = 0;
 tmpField = CreateBitField ();
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   node = (PHYLONODE*) tree -> root -> data;
   numNodes += node -> rec -> numVics;
 }
 nodeArray = g_ptr_array_sized_new (numNodes);
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   for (j = 0; j < tree -> nodeArray -> len; ++ j) {
     navNode = (GNode*) g_ptr_array_index (tree -> nodeArray, j);
     node = (PHYLONODE*) navNode -> data;
     if (node -> rec -> nodeFlag & VIC_NODE) {
       g_ptr_array_add (nodeArray, node);
       node -> alloSet = g_slice_alloc (sizeof (ALLOPATRY));
       node -> alloSet -> setA = CreateBitField ();
       node -> alloSet -> setB = CreateBitField ();
       CopyBitField (node -> alloSet -> setA, node -> rec -> setA -> rec -> inher);
       CopyBitField (node -> alloSet -> setB, node -> rec -> setB -> rec -> inher);
       BitFieldRemove (node -> alloSet -> setA, node -> rec -> setB -> rec -> inher);
       BitFieldRemove (node -> alloSet -> setB, node -> rec -> setA -> rec -> inher);
       node -> alloSet -> baseA = CreateBitField ();
       node -> alloSet -> baseB = CreateBitField ();
       CopyBitField (node -> alloSet -> baseA, node -> alloSet -> setA);
       CopyBitField (node -> alloSet -> baseB, node -> alloSet -> setB);
       node -> alloSet -> nodeList = NULL;
       node -> alloSet -> numNodes = 0;
       node -> alloSet -> triangles = NULL;
     }
   }
 }

 /* Compare nodes */
 for (j = 0; j < (nodeArray -> len - 1); ++ j) {
   node = (PHYLONODE*) g_ptr_array_index (nodeArray, j);
   for (k = j + 1; k < nodeArray -> len; ++ k) {
     compNode = (PHYLONODE*) g_ptr_array_index (nodeArray, k);
     if (IsNodeComp (node, compNode) == FALSE) continue;
     if (CompareSets (node -> alloSet -> setA, node -> alloSet -> setB, compNode -> alloSet -> setA, compNode -> alloSet -> setB)
                                     == FALSE) continue;
     nuList = InsertListElement (node -> alloSet -> nodeList, compNode);
     if (node -> alloSet -> nodeList == NULL)
       node -> alloSet -> nodeList = nuList;
     ++ node -> alloSet -> numNodes;
     nuList = InsertListElement (compNode -> alloSet -> nodeList, node);
     if (compNode -> alloSet -> nodeList == NULL)
       compNode -> alloSet -> nodeList = nuList;
     ++ compNode -> alloSet -> numNodes;
   }
 }
 /* Add distros */
 for (j = 0; j < (nodeArray -> len - 1); ++ j) {
   node = (PHYLONODE*) g_ptr_array_index (nodeArray, j);
   if (node -> alloSet -> numNodes == 0) continue;
   for (navComp = node -> alloSet -> nodeList; navComp != NULL; navComp = g_list_next (navComp)) {
     compNode = (PHYLONODE*) navComp -> data;
     if (BitFieldAnd (node -> alloSet -> baseA, compNode -> alloSet -> baseA)) {
       BitFieldORsum (node -> alloSet -> setA, compNode -> alloSet -> baseA);
       BitFieldORsum (node -> alloSet -> setB, compNode -> alloSet -> baseB);
     }
     else {
       BitFieldORsum (node -> alloSet -> setA, compNode -> alloSet -> baseB);
       BitFieldORsum (node -> alloSet -> setB, compNode -> alloSet -> baseA);
     }
   }
   CopyBitField (tmpField, node -> alloSet -> setA);
   BitFieldRemove (node -> alloSet -> setA, node -> alloSet -> setB);
   BitFieldRemove (node -> alloSet -> setB, tmpField);
 }
 g_slice_free1 (GETBITFIELDSIZE, tmpField);
 g_ptr_array_free (nodeArray, TRUE);
}
