#include "ghead.h"

GList* InsertListElement (GList* prev, gpointer data) {
 GList* nu;

 nu = g_list_append (NULL, data);
 if (prev != NULL) {
   nu -> prev = prev;
   nu -> next = prev -> next;
   if (prev -> next != NULL)
     prev -> next -> prev = nu;
   prev -> next = nu;
 }
 return nu;
}

void PutAtListEnd (GList* dest, GList* toPut) {
 GList* last;

 last = g_list_last (dest);
 toPut -> prev = last;
 if (last != NULL)
   last -> next = toPut;
}

void DestroyList (GList* toDel, gsize sz) {
 GList* nav;
 if (toDel == NULL) return;
 if (g_list_previous (toDel) != NULL)
   g_list_previous (toDel) -> next = NULL;
 for (nav = toDel; nav != NULL; nav = g_list_next (nav))
   g_slice_free1 (sz, nav -> data);
 g_list_free (toDel);
}

void DestroyLinkElement (GList* toDel, gsize sz) {
 if (toDel == NULL) return;
 if (g_list_previous (toDel) != NULL)
   g_list_previous (toDel) -> next = g_list_next (toDel);
 if (g_list_next (toDel) != NULL)
   g_list_next (toDel) -> prev = g_list_previous (toDel);
 toDel -> prev = toDel -> next = NULL;
 g_slice_free1 (sz, toDel -> data);
 g_list_free (toDel);
}

void DestroyLink (GList* toDel) {
 if (toDel == NULL) return;
 if (g_list_previous (toDel) != NULL)
   g_list_previous (toDel) -> next = g_list_next (toDel);
 if (g_list_next (toDel) != NULL)
   g_list_next (toDel) -> prev = g_list_previous (toDel);
 toDel -> prev = toDel -> next = NULL;
 g_list_free (toDel);
}

void AddChildren (GNode* toAdd, GNode* anc, GNode* lastSibling) {
 if (anc == NULL) return;
 toAdd -> parent = anc;
 if (lastSibling == NULL)
   toAdd -> parent -> children = toAdd;
 else
   lastSibling -> next = toAdd;
 toAdd -> prev = lastSibling;
}

GNode* GetNextNode (GNode* actNode) {
 GNode* navNode;

 if (actNode == NULL) return NULL;
 if (g_node_first_child (actNode) != NULL)
   return g_node_first_child (actNode);

 for (navNode = actNode;; navNode = navNode -> parent) {
   if (g_node_next_sibling (navNode) != NULL) return g_node_next_sibling (navNode);
   if (navNode -> parent == NULL) break;
 }
 return navNode;
}

GNode* GetNextTerm (GNode* actNode) {
 GNode* navNode;

 if (actNode == NULL) return NULL;
 navNode = actNode;
 /* Search the next available node, if there are no more nodes, returns NULL */
 if (g_node_first_child (actNode) == NULL) {
   for (; navNode != NULL; navNode = navNode -> parent) {
     if (g_node_next_sibling (navNode) != NULL) {
       navNode = g_node_next_sibling (navNode);
       break;
     }
   }
   if (navNode == NULL) return NULL;
 }
 for (; g_node_first_child (navNode) != NULL; navNode = g_node_first_child (navNode));
 return navNode;
}

GNode* GetPrevNode (GNode* actNode) {
 GNode* navNode;

 if (actNode == NULL) return NULL;
 if (g_node_prev_sibling (actNode) != NULL) return g_node_prev_sibling (actNode);
 if (actNode -> parent != NULL) return actNode -> parent;
 /* actNode is the root, so the "previous" node is the last children of the whole tree */
 for (navNode = actNode; ; navNode = g_node_last_child (navNode))
   if (g_node_last_child (navNode) == NULL) break;
 return navNode;
}

SFLAG IsDesc (GNode* probe, GNode* anc) {
 GNode* navNode;

 if ((probe == NULL) || (anc == NULL)) return FALSE;
 for (navNode = probe; navNode != NULL; navNode = navNode -> parent)
   if (navNode == anc) return TRUE;
 return FALSE;
}
