#include "voronoi.h"
#include <math.h>

static GList* beach = NULL;
static GList* circleQueue = NULL;

static void SiteEvent (SCOUNTER cell, SCOUNTER cRow, SCOUNTER cCol, SFLAG state);
static void CircleEvent (GPtrArray* triangleList);

static void CalculateCirclePoint (GList* navArc);

static SFLAG IntersectArc (GList* navArc, SREAL cRow, SREAL cCol);
static SREAL Intersect (BEACHARC* left, BEACHARC* right, SREAL cCol);

static GList* InsertEvent (CIRCEVENT* event);
static void RemoveEvent (GList* navEvent);
static SREAL GetEventCol ();

static void SearchNeighbors (GPtrArray* triangleList);

GPtrArray* MakeVoronoi (BITFIELD* setA, BITFIELD* setB) {
 SCOUNTER i, j;
 SCOUNTER actCell;
 SFLAG state;
 GPtrArray* triangleList;

 triangleList = g_ptr_array_new ();
 beach = NULL;

 /* Trips the grid, this is an intuitive implementation */
 for (j = 0; j < GETNUMCOLS; ++ j) {
   for (i = 0; i < GETNUMROWS; ++ i) {     /* Site events */
     actCell = GetCellIfObs (i, j);
     state = STATE_NOCELL;
     if (actCell != -1) {
       if (IsBitOn (setA, actCell) == TRUE)
         state = STATE_SETA;
       else if (IsBitOn (setB, actCell) == TRUE)
         state = STATE_SETB;
     }
     if (state != STATE_NOCELL) {
       while (circleQueue != NULL) {
         if (GetEventCol () <= j)
           CircleEvent (triangleList);
         else break;
       }
       SiteEvent (actCell, i, j, state);
     }
   }
 }
 /* Close remaining circle events */
 while (circleQueue != NULL)
   CircleEvent (triangleList);

 /* Release memory */
 DestroyList (beach, sizeof (BEACHARC));

 SearchNeighbors (triangleList);
 return triangleList;
}

void SiteEvent (SCOUNTER cell, SCOUNTER cRow, SCOUNTER cCol, SFLAG state) {
 BEACHARC* nuArc;
 BEACHARC* copyArc;
 BEACHARC* actArc;
 GList* nav;
 GList* pArc;

 nuArc = (BEACHARC*) g_slice_alloc (sizeof (BEACHARC));
 nuArc -> baseCell = cell;
 nuArc -> baseRow = cRow + 0.5;
 nuArc -> baseCol = cCol;
 nuArc -> baseState = state;
 nuArc -> event = NULL;
 nuArc -> lastCol = -1;
 nuArc -> lastLeft = -1;
 nuArc -> lastRight = -1;
 nuArc -> toDel = FALSE;
 if (beach == NULL) {
   beach = InsertListElement (NULL, nuArc);
   return;
 }
 for (nav = beach; nav != NULL; nav = g_list_next (nav)) {
   actArc = (BEACHARC*) nav -> data;
   if (actArc -> toDel == TRUE) continue;
   if (actArc -> lastCol == cCol) continue;
   if (IntersectArc (nav, nuArc -> baseRow, nuArc -> baseCol) != TRUE) {
     actArc -> lastCol = cCol;
     if (g_list_next (nav) == NULL) {
       pArc = InsertListElement (nav, nuArc);
       break;
     }
     continue;
   }
   copyArc = (BEACHARC*) g_slice_alloc (sizeof (BEACHARC));
   copyArc -> baseCell = actArc -> baseCell;
   copyArc -> baseRow = actArc -> baseRow;
   copyArc -> baseCol = actArc -> baseCol;
   copyArc -> baseState = actArc -> baseState;
   copyArc -> event = NULL;
   copyArc -> lastCol = -1;
   copyArc -> lastLeft = -1;
   copyArc -> lastRight = -1;
   copyArc -> toDel = FALSE;
   InsertListElement (nav, copyArc);
   pArc = InsertListElement (nav, nuArc);
   CalculateCirclePoint (nav);
   CalculateCirclePoint (pArc);
   CalculateCirclePoint (g_list_next (pArc));
   actArc -> lastCol = cCol;
   actArc -> lastRight = cRow + 0.5;
   return;
 }
}

void CircleEvent (GPtrArray* triangleList) {
 CIRCEVENT* event;
 BEACHARC* tmp1Arc;
 BEACHARC* tmp2Arc;
 BEACHARC* center;
 BEACHARC* left;
 BEACHARC* right;
 DELAUNAYTRIANGLE* triangle;
 GList* prevArc;
 GList* navArc;
 BEACHARC* actArc;

 /* Sort the points */
 event = (CIRCEVENT*) circleQueue -> data;
 navArc = event -> pArc;
 center = (BEACHARC*) navArc -> data;
 tmp1Arc = (BEACHARC*) g_list_previous (navArc) -> data;
 tmp2Arc = (BEACHARC*) g_list_next (navArc) -> data;
 /* Create the triangle and add it to list */
 if ((center -> baseState != tmp1Arc -> baseState) || (center -> baseState != tmp2Arc -> baseState)) {
   if (center -> baseCell > tmp1Arc -> baseCell) {
     if (center -> baseCell > tmp2Arc -> baseCell) {
       right = center;
       if (tmp1Arc -> baseCell > tmp2Arc -> baseCell) {
         left = tmp2Arc;
         center = tmp1Arc;
       }
       else {
         left = tmp1Arc;
         center = tmp2Arc;
       }
     }
     else {
       left = tmp1Arc;
       right = tmp2Arc;
     }
   }
   else {
     if (tmp1Arc -> baseCell > tmp2Arc -> baseCell) {
       right = tmp1Arc;
       if (center -> baseCell > tmp2Arc -> baseCell)
         left = tmp2Arc;
       else {
         left = center;
         center = tmp2Arc;
       }
     }
     else {
       right = tmp2Arc;
       left = center;
       center = tmp1Arc;
     }
   }
   triangle = g_slice_alloc (sizeof (DELAUNAYTRIANGLE));
   triangle -> cellA = left -> baseCell;
   triangle -> stateA = left -> baseState;
   triangle -> cellB = center -> baseCell;
   triangle -> stateB = center -> baseState;
   triangle -> cellC = right -> baseCell;
   triangle -> stateC = right -> baseState;
   triangle -> cX = event -> cX;
   triangle -> cY = event -> cY;
   triangle -> abNeighbor = 0;
   triangle -> acNeighbor = 0;
   triangle -> bcNeighbor = 0;
   g_ptr_array_add (triangleList, triangle);
 }
 prevArc = g_list_previous (navArc);
 /* Removes the arc */
 RemoveEvent (circleQueue);
 DestroyLinkElement (navArc, sizeof (BEACHARC));
 CalculateCirclePoint (prevArc);
 CalculateCirclePoint (g_list_next (prevArc));
 /* Reset the beach last update */
 for (navArc = prevArc; navArc != NULL; navArc = g_list_next (navArc)) {
   actArc = (BEACHARC*) navArc -> data;
   if (actArc -> lastCol == -1) break;
   actArc -> lastCol = -1;
 }
}

void DestroyTriangleList (GPtrArray* triangleList) {
 SCOUNTER i;
 DELAUNAYTRIANGLE* toDel;

 for (i = 0; i < triangleList -> len; ++ i) {
   toDel = (DELAUNAYTRIANGLE*) g_ptr_array_index (triangleList, i);
   g_slice_free1 (sizeof (DELAUNAYTRIANGLE), toDel);
 }
 g_ptr_array_free (triangleList, TRUE);
}

void CalculateCirclePoint (GList* navArc) {
 SREAL aX, aY, bX, bY, aN, bN, deno, cX, cY, r, circleCol;
 BEACHARC* left;
 BEACHARC* actArc;
 BEACHARC* right;
 GList* prevArc;
 GList* nextArc;
 CIRCEVENT* event;
 CIRCEVENT* neigEvent;

 prevArc = g_list_previous (navArc);
 nextArc = g_list_next (navArc);
 actArc = (BEACHARC*) navArc -> data;
 RemoveEvent (actArc -> event);
 if ((prevArc != NULL) && (nextArc != NULL)) {
   left = (BEACHARC*) prevArc -> data;
   right = (BEACHARC*) nextArc -> data;
   if (left -> baseCell == right -> baseCell) return;
   aY = left -> baseRow - actArc -> baseRow;
   aX = left -> baseCol - actArc -> baseCol;
   bY = right -> baseRow - actArc -> baseRow;
   bX = right -> baseCol - actArc -> baseCol;
   deno = 2 * ((aX * bY) - (aY * bX));
   if (deno <= 0.0000001) return;               /* Colineal or unsorted points */
   aN = (aX * aX) + (aY * aY);
   bN = (bX * bX) + (bY * bY);
   cX = ((aN * bY) - (bN * aY)) / deno;
   cY = ((bN * aX) - (aN * bX)) / deno;
   r = sqrt (((cX - aX) * (cX - aX)) + ((cY - aY) * (cY - aY)));
   circleCol = cX + r + actArc -> baseCol;
   /* Check if any neighbor has a circle event before the detected circle event */
   if (left -> event != NULL) {
     neigEvent = (CIRCEVENT*) left -> event -> data;
     if (neigEvent -> circleCol < circleCol) return;
   }
   if (right -> event != NULL) {
     neigEvent = (CIRCEVENT*) right -> event -> data;
     if (neigEvent -> circleCol < circleCol) return;
   }
   event = g_slice_alloc (sizeof (CIRCEVENT));
   event -> pArc = navArc;
   event -> circleCol = circleCol;
   event -> cX = cX + actArc -> baseCol;
   event -> cY = cY + actArc -> baseRow;
   actArc -> event = InsertEvent (event);
 }
}

SFLAG IntersectArc (GList* navArc, SREAL cRow, SREAL cCol) {
 GList* navLeft;
 GList* navRight;
 BEACHARC* left;
 BEACHARC* right;
 BEACHARC* actArc;

 actArc = (BEACHARC*) navArc -> data;
 if (actArc -> baseCol == cCol) {
   actArc -> lastLeft = actArc -> lastRight = actArc -> baseRow;
   return FALSE;
 }
 navLeft = g_list_previous (navArc);
 navRight = g_list_next (navArc);
 actArc -> lastLeft = 0;
 if (navLeft != NULL) {
   left = (BEACHARC*) navLeft -> data;
   if (left -> toDel != TRUE)
     actArc -> lastLeft = left -> lastRight;
 }
 if (navRight != NULL) {
   right = (BEACHARC*) navRight -> data;
   actArc -> lastRight = Intersect (actArc, right, cCol);
 }
 else
   actArc -> lastRight = GETNUMROWS;
 if (actArc -> lastRight < actArc -> lastLeft) {
   /* Arcs outside the frame are not revisited */
   actArc -> toDel = TRUE;
   actArc -> lastRight = actArc -> lastLeft;
   return FALSE;
 }
 if ((cRow >= actArc -> lastLeft) && (cRow <= actArc -> lastRight))
   return TRUE;
 return FALSE;
}

SREAL Intersect (BEACHARC* left, BEACHARC* right, SREAL cCol) {
 SREAL distLeft, distRight, a, b, c;
 SREAL rootTerm;
 SREAL row1, row2;
 SREAL toRet;

 if (left -> baseCol == right -> baseCol) {
   distRight = (right -> baseRow - left -> baseRow) / 2;
   return (left -> baseRow) + distRight;
 }
 distLeft = (SREAL) (left -> baseCol - cCol);
 distRight = (SREAL) (right -> baseCol - cCol);
 if (distLeft == 0) return left -> baseRow;
 if (distRight == 0) return right -> baseRow;
 a = (1.00 / (2 * distLeft)) - (1.00 / (2 * distRight));
 b = (right -> baseRow / distRight) - (left -> baseRow / distLeft);
 c = (left -> baseRow * left -> baseRow / (2 * distLeft)) - (right -> baseRow * right -> baseRow / (2 * distRight))
     + left -> baseCol - right -> baseCol + (distLeft / 2) - (distRight / 2);
 rootTerm = sqrt ((b * b) - (4 * a * c));
 row1 = (- b + rootTerm) / (2 * a);
 row2 = (- b - rootTerm) / (2 * a);
 /* Look for the arc configuration */
 if (left -> baseCol < right -> baseCol)
   toRet = (row1 < row2) ? row1 : row2;
 else
   toRet = (row1 > row2) ? row1 : row2;
 return toRet;
}

GList* InsertEvent (CIRCEVENT* event) {
 CIRCEVENT* actEvent;
 GList* navEvent;
 GList* toRet;

 if (circleQueue == NULL) {
   circleQueue = toRet = InsertListElement (NULL, event);
   return toRet;
 }
 toRet = NULL;
 for (navEvent = circleQueue; ; navEvent = g_list_next (navEvent)) {
   actEvent = (CIRCEVENT*) navEvent -> data;
   if (actEvent -> circleCol > event -> circleCol) {
     if (navEvent == circleQueue) {
       toRet = InsertListElement (NULL, event);
       toRet -> next = circleQueue;
       circleQueue -> prev = toRet;
       circleQueue = toRet;
     }
     else
       toRet = InsertListElement (g_list_previous (navEvent), event);
     break;
   }
   else if (g_list_next (navEvent) == NULL) {
     toRet = InsertListElement (navEvent, event);
     break;
   }
 }
 return toRet;
}

void RemoveEvent (GList* navEvent) {
 BEACHARC* actArc;
 CIRCEVENT* actEvent;

 if (navEvent == NULL) return;
 actEvent = (CIRCEVENT*) navEvent -> data;
 actArc = (BEACHARC*) actEvent -> pArc -> data;
 actArc -> event = NULL;
 if (navEvent == circleQueue)
   circleQueue = g_list_next (circleQueue);
 DestroyLinkElement (navEvent, sizeof (CIRCEVENT));
}

SREAL GetEventCol () {
 CIRCEVENT* actEvent;

 actEvent = (CIRCEVENT*) circleQueue -> data;
 return actEvent -> circleCol;
}

void SearchNeighbors (GPtrArray* triangleList) {
 SCOUNTER i, j;
 DELAUNAYTRIANGLE* triangleA;
 DELAUNAYTRIANGLE* triangleB;
 SCOUNTER neighbors;

 if (triangleList -> len == 0) return;
 for (i = 0; i < (triangleList -> len - 1); ++ i) {
   triangleA = (DELAUNAYTRIANGLE*) g_ptr_array_index (triangleList, i);
   neighbors = 0;
   if (triangleA -> abNeighbor != 0)
     ++ neighbors;
   if (triangleA -> acNeighbor != 0)
     ++ neighbors;
   if (triangleA -> bcNeighbor != 0)
     ++ neighbors;
   for (j = i + 1; j < triangleList -> len; ++ j) {
     if (neighbors == 3) break;
     triangleB = (DELAUNAYTRIANGLE*) g_ptr_array_index (triangleList, j);
     if (triangleA -> abNeighbor == 0) {
       if (triangleA -> cellA == triangleB -> cellA) {
         if (triangleA -> cellB == triangleB -> cellB) {
           triangleA -> abNeighbor = j;
           triangleB -> abNeighbor = i;
           ++ neighbors;
           continue;
         }
         else if (triangleA -> cellB == triangleB -> cellC) {
           triangleA -> abNeighbor = j;
           triangleB -> acNeighbor = i;
           ++ neighbors;
           continue;
         }
       }
       else if (triangleA -> cellA == triangleB -> cellB) {
         if (triangleA -> cellB == triangleB -> cellC) {
           triangleA -> abNeighbor = j;
           triangleB -> bcNeighbor = i;
           ++ neighbors;
           continue;
         }
       }
     }
     if (triangleA -> acNeighbor == 0) {
       if (triangleA -> cellA == triangleB -> cellA) {
         if (triangleA -> cellC == triangleB -> cellB) {
           triangleA -> acNeighbor = j;
           triangleB -> abNeighbor = i;
           ++ neighbors;
           continue;
         }
         else if (triangleA -> cellC == triangleB -> cellC) {
           triangleA -> acNeighbor = j;
           triangleB -> acNeighbor = i;
           ++ neighbors;
           continue;
         }
       }
       else if (triangleA -> cellA == triangleB -> cellB) {
         if (triangleA -> cellC == triangleB -> cellC) {
           triangleA -> acNeighbor = j;
           triangleB -> bcNeighbor = i;
           ++ neighbors;
           continue;
         }
       }
     }
     if (triangleA -> bcNeighbor == 0) {
       if (triangleA -> cellB == triangleB -> cellA) {
         if (triangleA -> cellC == triangleB -> cellB) {
           triangleA -> bcNeighbor = j;
           triangleB -> abNeighbor = i;
           ++ neighbors;
           continue;
         }
         else if (triangleA -> cellC == triangleB -> cellC) {
           triangleA -> bcNeighbor = j;
           triangleB -> acNeighbor = i;
           ++ neighbors;
           continue;
         }
       }
       else if (triangleA -> cellB == triangleB -> cellB) {
         if (triangleA -> cellC == triangleB -> cellC) {
           triangleA -> bcNeighbor = j;
           triangleB -> bcNeighbor = i;
           ++ neighbors;
           continue;
         }
       }
     }
   }
 }
}
